/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
/*
 * senddump.c - runs on machine being dumped
 */
#include "amanda.h"
#include "stream.h"

#define BUFFER_SIZE	32*1024

/*
 * If we don't have the new-style wait access functions, use our own,
 * compatible with old-style BSD systems at least.  Note that we don't
 * care about the case w_stopval == WSTOPPED since we don't ask to see
 * stopped processes, so should never get them from wait.
 */
#ifndef WEXITSTATUS
#   define WEXITSTATUS(r)	(((union wait *) &(r))->w_retcode)
#   define WTERMSIG(r)		(((union wait *) &(r))->w_termsig)

#   undef IFSIGNALED
#   define IFSIGNALED(r)	(((uniion wait *) &(r))->w_termsig != 0)
#endif

char *pname = "senddump";
struct timeval start_time;
int comppid, dumppid;
char thiserr[80], errorstr[256];

double tv2f(tv)
struct timeval tv;
/*
 * Converts a timeval into a double giving the time
 * in seconds.
 */
{
    return tv.tv_sec + tv.tv_usec / 1e6;
}

char *childstr(pid)
int pid;
/*
 * Returns a string for a child process.  Checks the saved dump and
 * compress pids to see which it is.
 */
{
    if(pid == dumppid) return "dump";
    if(pid == comppid) return "compress";
    return "unknown";
}


#ifdef INSTRUMENTATION

void print_childstats(pid, usage)
int pid;
struct rusage usage;
/*
 * Prints all the interesting statistics for a child process.
 */
{
    double utime, stime;

    utime = tv2f(usage.ru_utime);
    stime = tv2f(usage.ru_stime);

    fprintf(stderr,
  "dumpstats: %s times %1.2fu %1.2fs i/o %d %d r/s %d %d sig %d i/vcs %d %d\n",
	    childstr(pid), utime, stime, 
	    usage.ru_inblock, usage.ru_oublock,
	    usage.ru_msgrcv,  usage.ru_msgsnd,
	    usage.ru_nsignals,
	    usage.ru_nvcsw, usage.ru_nivcsw);
}


void print_totalstats()
/*
 * Prints the overall timing statistics.
 */
{
    struct timeval end_time;
    struct rusage self, child;
    double rtime, utime, stime;

    getrusage(RUSAGE_SELF, &self);
    getrusage(RUSAGE_CHILDREN, &child);
    gettimeofday(&end_time, NULL);

    rtime = tv2f(end_time) - tv2f(start_time);
    utime = tv2f(self.ru_utime) + tv2f(child.ru_utime);
    stime = tv2f(self.ru_stime) + tv2f(child.ru_stime);

    fprintf(stderr, "dumpstats: slave total times %1.2fr %1.2fu %1.2fs\n",
	    rtime, utime, stime);
}

#endif


int check_status(pid, w)
int pid, w;
/*
 * Determine if the child return status really indicates an error.
 * If so, add the error message to the error string; more than one
 * child can have an error.
 */
{
    char *str;
    int ret, sig;

    if(WIFSIGNALED(w))	ret = 0, sig = WTERMSIG(w);
    else sig = 0, ret = WEXITSTATUS(w);

    str = childstr(pid);

    if(pid == comppid) 	/* compress returns 2 sometimes; it's ok */
	if(ret == 2) return 0;

#ifdef DUMP_RETURNS_1
    if(pid == dumppid) /* Ultrix dump returns 1 sometimes; it's ok too */
        if(ret == 1) return 0;
#endif

    if(ret == 0) sprintf(thiserr, "%s got signal %d", str, sig);
    else sprintf(thiserr, "%s returned %d", str, ret);

    if(*errorstr != '\0') strcat(errorstr, ", ");
    strcat(errorstr, thiserr);
    return 1;
}


void write_tapeheader(host, disk, level, compress, datestamp, outf)
char *host, *disk, *datestamp;
int outf, compress, level;
/*
 * writes an Amanda tape header onto the output file.
 */
{
    char line[128], buffer[BUFFER_SIZE];
    int len, rc;

    sprintf(buffer, "AMANDA: FILE %s %s %s lev %d comp %c\n",
            datestamp, host, disk, level, 
	    compress? 'C': 'N');
    

    strcat(buffer,"To restore, position tape at start of file and run:\n");

    sprintf(line,"\tdd if=<tape> bs=%dk skip=1 |%s restore ...f -\n\014\n",
            BUFFER_SIZE/1024, compress? " zcat |" : "");
    strcat(buffer, line);

    len = strlen(buffer);
    memset(buffer+len, '\0', BUFFER_SIZE-len);
    if((rc = write(outf, buffer, BUFFER_SIZE)) < BUFFER_SIZE)
	error("error [write header: %s]", strerror(errno));
}


int compress, no_record;

void handle_options(str)
char *str;
{
    /* only two options, no need to get fancy */
    compress = strstr(str, "compress") != NULL;
    no_record = strstr(str, "no-record") != NULL;
}

#ifndef BSD_SECURITY
/* stopgap security method, for testing */

char *okhosts[] = TRUSTED_HOSTS;
#define NUM_TRUSTED_HOSTS (sizeof(okhosts)/sizeof(char *))

int lookup_host(hostname)
char *hostname;
{
    char **ok;

    for(ok = okhosts; ok < okhosts + NUM_TRUSTED_HOSTS; ok++)
	if(!strcmp(hostname, *ok)) return 1;
    return 0;
}
#endif

void main(argc, argv)
int argc;
char **argv;
{
    int rc, level, remote_port, sockbuf;
    int datafd, mesgfd, dumpout, pipefd[2], nullfd, wpid, goterror;
    int retstat;
    struct rusage child_rusage;
    struct sockaddr_in peername;
    struct hostent *hp, *gethostbyaddr();
    int namelen, domlen;
    char program[80], options[80], disk[80], host[80], dumpkeys[10];
    char device[80], remotehost[80], datestamp[80], *domain;
    char remoteuser[80], localuser[80], *pwname;
    struct passwd *pwptr;
    int myuid;

    erroutput_type = (ERR_INTERACTIVE|ERR_SYSLOG);

    mesgfd = 1;		/* already set up by inetd */
    dup2(mesgfd,2);	/* set up stderr */

    gettimeofday(&start_time, NULL);

    /* who am I? */

    if(gethostname(host, sizeof host) == -1)
	error("error [gethostname: %s]", strerror(errno));
    domain = strchr(host, '.');
    if(domain) {
	*domain++ = '\0';
	domlen = strlen(domain);
    }
    else domlen = 0;

#ifdef FORCE_DUMPUSER

    /* we'd rather not run as root */

    if(geteuid() == 0) {
	pwname = SLAVE_LOGIN;
	if((pwptr = getpwnam(pwname)) == NULL)
	    error("error [cannot find user %s in passwd file]\n", pwname);

	initgroups(pwname, pwptr->pw_gid);
	setgid(pwptr->pw_gid);
	setuid(pwptr->pw_uid);
    }

#endif

    /* who am I talking to? */

    namelen = sizeof(peername);
    if(getpeername(0, (struct sockaddr *)&peername, &namelen) == -1)
       error("error [getpeername: %s]", strerror(errno));

    hp = gethostbyaddr((char *)&peername.sin_addr, 
		       sizeof(peername.sin_addr), AF_INET);
    if(hp == NULL) error("error [gethostbyaddr: %s]", strerror(errno));
    strcpy(remotehost, hp->h_name);

    /* get arguments from stdin */
    /* !!!! should timeout here? */
    rc = scanf("%d %s %s %d %d %s %s %[^\n]", &remote_port, program, disk,
	       &level, &sockbuf, datestamp, remoteuser, options);
    if(rc != 8) error("error [incorrect usage]");

    /*
     * Go ahead and do connect before security check.  Dumper is waiting
     * for this and won't even look at an error message unless it gets
     * the connection.  If the security check fails we'll close the connection
     * right away without sending any filesystem data, but then at least
     * the operators will see the error message in their reports rather than
     * just "accept timed out".
     *
     * After all, the most common cause of problems here is going to be
     * operators setting things up incorrectly, not some hostile attack.
     */

    datafd = stream_client(remotehost, remote_port, sockbuf, DEFAULT_SIZE);
    if(datafd == -1)
	error("error [data stream connect to %s port %s: %s]",
	      remotehost, remote_port, strerror(errno));

    if(strcmp(program, "DUMP"))
	error("error [cannot handle backup type \"%s\"]", program);

    handle_options(options);

    printf("senddump: start %s %s level %d to %s\n", host, disk,
	   level, remotehost); fflush(stdout);

#ifdef BSD_SECURITY

    if(remote_port >= IPPORT_RESERVED)
	error("error [attempt from %s, port %d not secure]", 
	      remotehost, peername.sin_port);

    myuid = getuid();
    if((pwptr = getpwuid(myuid)) == NULL)
	error("error [getpwuid(%d) fails]", myuid);

    strcpy(localuser, pwptr->pw_name);

    /*
     * note that some versions of ruserok (eg SunOS 3.2) look in
     * "./.rhosts" rather than "~localuser/.rhosts", so we have to
     * chdir ourselves.  Sigh.
     *
     * And, beleive it or not, some ruserok()'s try an initgroup just
     * for the hell of it.  Since we probably aren't root at this point
     * it'll fail, and initgroup "helpfully" will blatt "Setgroups: Not owner"
     * into our dump output even though the initgroup failure is not a problem
     * and is expected.  Thanks a lot.  Not.
     */
    chdir(pwptr->pw_dir);	/* pamper braindead ruserok's */
    close(2);			/*  " */

    if(ruserok(remotehost, myuid == 0, remoteuser, localuser) == -1) {
	dup2(1,2);
        error("error [access as %s not allowed from %s@%s]",
		localuser, remoteuser, remotehost);
    }

    dup2(1,2);
    chdir("/tmp");	/* now go someplace where I can drop core :-) */

#else

    /* a somewhat lame security check */

    if(!lookup_host(remotehost))
	error("error [%s is an unauthorized host]", remotehost);

#endif

    write_tapeheader(host, disk, level, compress, datestamp, datafd);

    if((nullfd = open("/dev/null", O_RDWR)) == -1)
	error("error [open /dev/null: %s]", strerror(errno));

    if(!compress) {
	dumpout = datafd;
	comppid = -1;
    } else {
	pipe(pipefd);
	dumpout = pipefd[1];

	switch(comppid = fork()) {
	case -1: error("error [fork compress: %s]", strerror(errno));
	default:	/* parent process */
	    close(pipefd[0]);	/* close input side of pipe */
	    close(datafd);	/* and compress output */
	    break;
	case 0:		/* child process */
	    close(pipefd[1]);	/* close output side of pipe */

	    if(dup2(pipefd[0], 0) == -1)
		error("error [dup2 pipe: %s]", strerror(errno));
	    if(dup2(datafd, 1) == -1)
		error("error [dup2 data: %s]", strerror(errno));

	    execl(COMPRESS, "compress", 0);
	    error("error [exec compress: %s]", strerror(errno));
	}
    }

    switch(dumppid = fork()) {
    case -1: error("error [fork dump: %s]", strerror(errno));
    default: break;
    case 0:	/* child process */
	if(dup2(nullfd, 0) == -1)
	    error("error [dup2 dumpin: %s]", strerror(errno));
	if(dup2(dumpout, 1) == -1)
	    error("error [dup2 dumpout: %s]", strerror(errno));

#ifndef AIX_BACKUP
	/* normal dump */
	sprintf(dumpkeys, "%d%ssf", level, no_record ? "" : "u");
	sprintf(device, "%s/%s", DEVDIR, disk);

	execl(DUMP, "dump", dumpkeys, "100000", "-", device, 0);
	error("error [exec dump: %s]", strerror(errno));
#else
	/* AIX backup program */
	/* XXX - we are going to need to generalize this */

	sprintf(dumpkeys, "-%d%sf", level, no_record ? "" : "u");
	sprintf(device, "%s/%s", DEVDIR, disk);

	execl(DUMP, "backup", dumpkeys, "-", device, 0);
	error("error [exec dump: %s]", strerror(errno));
#endif
    }

    close(dumpout);

    goterror = 0;
    *errorstr = '\0';

    while((wpid = wait(&retstat)) != -1) {
	/* we know that it exited, so we don't have to check WIFEXITED */
	if((WIFSIGNALED(retstat) || WEXITSTATUS(retstat)) &&
	   check_status(wpid, retstat))
	    goterror = 1;
    }

    if(goterror) fprintf(stderr, "senddump: error [%s]\n", errorstr);
    else fputs("senddump: end\n", stderr);
    exit(0);
}
